/* *********************************************************************/
/**
* Copyright (c) 2017 Mimer Information Technology
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
* copies of the Software, and to permit persons to whom the Software is
* furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* See license for more details.
 */
/* *********************************************************************/

/* *********************************************************************/
/**
 *   @brief   C interface for accessing Mimer from Python
 *
 *   @file
 *
 *   @par   File information
 *          - Mimer Information Technology
 *          - System:    Mimer SQL
 *          - Module:    DRIVERS
 *          - Submodule: MIMERPY
 */
/* *********************************************************************/

#include "cursor.h"
#include <stdio.h>
#include <stdlib.h>

static PyObject* mimerBeginStatementC(PyObject* self, PyObject* args)
/* *********************************************************************/
/**
 *  @brief      Prepares an SQL statement for execution.
 *
 *  @author     Erik Gunne & Magdalena Boström
 *  @date       2017-06-16
 *
 *  @param[in]     self       Pointer to python object that function is
 *                            called from.
 *  @param[in] args           Arguments.
 *                 args 1:    Sessionhandle.
 *                 args 2:    SQL statement to execute.
 *                 args 3:    Specifies a cursor being scrollable, or forward
 *                            only. 0 indicate forward only, 1 scrollable.
 *  @returns    Return code 0 = OK, < 0 if error.
 *
 *  @remarks    Between the preparation and the time of execution, the
 *              the application may supply any number of parameters to be
 *              used when executing the statement.
 *  @remarks    The same statement does not nsessioneed to prepared again, even
 *              if it is executed multiple times.
 *  @remarks    There is no need to prepare all statements at the start of
 *              the application.
 */
/* *********************************************************************/
{
  void *session;
  char* query;
  int rc, opt;
  MimerStatement statement;

  if(!PyArg_ParseTuple(args, "Lsi", &session, &query, &opt)){
    return NULL;
  }

  rc = MimerBeginStatementC(session, query, opt, &statement);

  return Py_BuildValue("iL", rc, (int64_t)statement);
}

static PyObject* mimerEndStatement(PyObject* self, PyObject* args)
/* *********************************************************************/
/**
 *  @brief      Closes a prepared statement.
 *
 *  @author     Erik Gunne & Magdalena Boström
 *  @date       2017-06-16
 *
 *  @param[in]     self       Pointer to python object that function is
 *                            called from.
 *  @param[in]     args       Arguments.
 *                 args 1:    Statementhandle.
 *  @returns    Return code 0 = OK, < 0 if error.
 *
 *  @remarks    If there is an open cursor on this statement, it is
 *              automatically closed.
 */
/* *********************************************************************/
{
  MimerStatement statement;
  int rc;

  if(!PyArg_ParseTuple(args, "L", &statement)){
    return NULL;
  }

  rc = MimerEndStatement(&statement);
  return Py_BuildValue("i", rc);
}

static PyObject* mimerOpenCursor(PyObject* self, PyObject* args)
/* *********************************************************************/
/**
 *  @brief      Opens a cursor to be used when reading a result set.
 *
 *  @author     Erik Gunne & Magdalena Boström
 *  @date       2017-06-16
 *
 *  @param[in]     self       Pointer to python object that function is
 *                            called from.
 *  @param[in]     args       Arguments.
 *                 args 1:    Statementhandle.
 *  @returns    Return code 0 = OK, < 0 if error.
 *
 *  @remarks    Result sets are produced either by SELECT queries, or by
 *              calls to result set procedures.
 *  @remarks    All input parameters must be set prior to calling
 *              mimerOpenCursor, or else an error will occur.
 *  @remarks    If this routine returns successfully, the cursor is opened
 *              and positioned before the first row.
 *  @remarks    The first row can then be fetched by calling mimerFetch.
 *              After that column data may be retrieved using data
 *              output routines.
 *  @remarks    If the statement does not return a result set, this
 *              routine will return with a failure.
 */
/* *********************************************************************/
{
  int rc;
  void *statement;

  if(!PyArg_ParseTuple(args, "L", &statement)){
    return NULL;
  }

  rc = MimerOpenCursor(statement);
  return Py_BuildValue("i", rc);
}

static PyObject* mimerCloseCursor(PyObject* self, PyObject* args)
/* *********************************************************************/
/**
 *  @brief      Closes an open cursor.
 *
 *  @author     Erik Gunne & Magdalena Boström
 *  @date       2017-06-16
 *
 *  @param[in]     self       Pointer to python object that function is
 *                            called from.
 *  @param[in]     args       Arguments.
 *                 args 1:    Statementhandle.
 *  @returns    Return code 0 = OK, < 0 if error.
 *
 *  @remarks    When the cursor is closed, all resources held by the
 *              result set are released.
 *  @remarks    After the cursor is closed, the cursor may be reopened
 *              using a call to mimerOpenCursor or the statement should
 *              be released using mimerEndStatement.
 *  @remarks    Before the cursor is reopened, a new set of parameters
 *              may be supplied using any of the data input functions.
 */
/* *********************************************************************/
{
  int rc;
  void *statement;

  if(!PyArg_ParseTuple(args, "L", &statement)){
    return NULL;
  }

  rc = MimerCloseCursor(statement);
  return Py_BuildValue("i", rc);
}

static PyObject* mimerExecuteStatementC(PyObject* self, PyObject* args)
/* *********************************************************************/
/**
 *  @brief      Executes a statment directly without parameters.
 *
 *  @author     Erik Gunne & Magdalena Boström
 *  @date       2017-06-16
 *
 *  @param[in]    self       Pointer to python object that function is
 *                           called from.
 *  @param[in]    args       Arguments.
 *                args 1:    Statementhandle.
 *                args 2:    SQL statement to execute.
 *  @returns    Return code 0 = OK, < 0 if error, > 0 indicates an update
 *              row count returned by the database.
 *
 *  @remarks    This routine is mainly intended for data definition
 *              statements but can also be used for regular INSERT,
 *              UPDATE and DELETE statements with no input or output
 *              parameters.
 */
/* *********************************************************************/
{
  void *session;
  char *query;
  int rc;

  if(!PyArg_ParseTuple(args, "Ls", &session, &query)){
    return NULL;
  }

  rc = MimerExecuteStatementC((MimerSession)session, query);
  return Py_BuildValue("i", rc);
}

static PyObject* mimerExecute(PyObject* self, PyObject* args)
/* *********************************************************************/
/**
 *  @brief      Execute a statement that does not return a result set.
 *
 *  @author     Erik Gunne & Magdalena Boström
 *  @date       2017-06-16
 *
 *  @param[in]    self        Pointer to python object that function is
 *                            called from.
 *  @param[in]    args        Arguments.
 *                args 1:     Statementhandle.
 *  @returns    Return code 0 = OK, < 0 if error, > 0 indicates an update
 *              row count returned by the database.
 *
 *  @remarks    This routine is mainly intended for data definition
 *              statements but can also be used for regular INSERT,
 *              UPDATE and DELETE statements with no input or output
 *              parameters.
 */
/* *********************************************************************/
{
  void *statement;
  int rc;

  if(!PyArg_ParseTuple(args, "L", &statement)){
    return NULL;
  }

  rc = MimerExecute(statement);
  return Py_BuildValue("i", rc);
}

static PyObject* mimerParameterCount(PyObject* self, PyObject* args)
/* *********************************************************************/
/**
 *  @brief      Obtains the number of parameters of a statement.
 *
 *  @author     Erik Gunne & Magdalena Boström
 *  @date       2017-06-16
 *
 *  @param[in]    self        Pointer to python object that function is
 *                            called from.
 *  @param[in]    args        Arguments.
 *                args 1:     Statementhandle.
 *  @returns    Return code >= 0 the number of parameters, < 0 if error.
 */
/* *********************************************************************/
{
  void *statement;
  int rc;

  if(!PyArg_ParseTuple(args, "L", &statement)){
    return NULL;
  }

  rc = MimerParameterCount(statement);
  return Py_BuildValue("i", rc);
}

static PyObject* mimerParameterType(PyObject* self, PyObject* args)
/* *********************************************************************/
/**
 *  @brief      Obtains the type of a parameter.
 *
 *  @author     Erik Gunne & Magdalena Boström
 *  @date       2017-06-16
 *
 *  @param[in]    self        Pointer to python object that function is
 *                            called from.
 *  @param[in]    args        Arguments.
 *                args 1:     Statementhandle.
 *                args 2:     Parameter number.
 *  @returns    Return code > 0 indicates parameter type, < 0 if error.
 */
/* *********************************************************************/
{
  void *statement;
  int rc, parameter_number;

  if(!PyArg_ParseTuple(args, "Li", &statement, &parameter_number)){
    return NULL;
  }

  rc = MimerParameterType(statement, parameter_number);
  return Py_BuildValue("i", rc);
}

static PyObject* mimerColumnCount(PyObject* self, PyObject* args)
/* *********************************************************************/
/**
 *  @brief      Obtains the number of columns in a result set.
 *
 *  @author     Erik Gunne & Magdalena Boström
 *  @date       2017-06-16
 *
 *  @param[in]    self        Pointer to python object that function is
 *                            called from.
 *  @param[in]    args        Arguments.
 *                args 1:     Statementhandle.
 *  @returns    Return code >= 0 number of result set columns, < 0 if error.
 */
/* *********************************************************************/
{
  void *statement;
  int rc;

  if(!PyArg_ParseTuple(args, "L", &statement)){
    return NULL;
  }

  rc = MimerColumnCount(statement);
  return Py_BuildValue("i", rc);
}

static PyObject* mimerColumnType(PyObject* self, PyObject* args)
/* *********************************************************************/
/**
 *  @brief      Returns the type of a column.
 *
 *  @author     Erik Gunne & Magdalena Boström
 *  @date       2017-06-16
 *
 *  @param[in]    self        Pointer to python object that function is
 *                            called from.
 *  @param[in]    args        Arguments.
 *                args 1:     Statementhandle.
 *                args 2:     Column number.
 *  @returns    Return code > 0 indicates column type, < 0 if error.
 */
/* *********************************************************************/
{
  void *statement;
  int rc, column_number;

  if(!PyArg_ParseTuple(args, "Li", &statement, &column_number)){
    return NULL;
  }

  rc = MimerColumnType(statement, column_number);
  return Py_BuildValue("i", rc);
}

static PyObject* mimerFetch(PyObject* self, PyObject* args)
/* *********************************************************************/
/**
 *  @brief      Advances to the next row of a result set.
 *
 *  @author     Erik Gunne & Magdalena Boström
 *  @date       2017-06-16
 *
 *  @param[in]    self        Pointer to python object that function is
 *                            called from.
 *  @param[in]    args        Arguments.
 *                args 1:     Statementhandle.
 *  @returns    Return code 0 = OK, < 0 if error.
 *
 *  @remarks    If the routine returns successfully, the current row has
 *              been advanced one row down the result set. Column data
 *              may be retrieved using data output routines.
 */
/* *********************************************************************/
{
  void *statement;
  int rc;

  if(!PyArg_ParseTuple(args, "L", &statement)){
    return NULL;
  }

  rc = MimerFetch(statement);
  return Py_BuildValue("i", rc);
}

static PyObject* mimerGetInt32(PyObject* self, PyObject* args)
/* *********************************************************************/
/**
 *  @brief      Gets integer data from a result set or output parameter.
 *
 *  @author     Erik Gunne & Magdalena Boström
 *  @date       2017-06-16
 *
 *  @param[in]    self        Pointer to python object that function is
 *                            called from.
 *  @param[in]    args        Arguments.
 *                args 1:     Statementhandle.
 *                args 2:     Parameter of column number to get data from.
 *  @returns   Tuple.
 *             Tuple[0]: Return code 0 = OK, < 0 if error.
 *             Tuple[1]: The value of the integer.
 *
 *  @remarks   Only the database type INTEGER, INTEGER(n), BIGINT and SMALLINT
 *             may be retrieved using this call.
 */
/* *********************************************************************/
{
  void *statement;
  int rc, return_value, column_number;

  if(!PyArg_ParseTuple(args, "Li", &statement, &column_number)){
    return NULL;
  }

  if(MimerIsNull(statement, column_number) > 0){
    return Py_BuildValue("is", 0, NULL);
  }

  rc = MimerGetInt32(statement, column_number, &return_value);
  return Py_BuildValue("ii", rc, return_value);
}

static PyObject* mimerGetInt64(PyObject* self, PyObject* args)
/* *********************************************************************/
/**
 *  @brief      Gets integer data from a result set or output parameter.
 *
 *  @author     Erik Gunne & Magdalena Boström
 *  @date       2017-06-16
 *
 *  @param[in]    self        Pointer to python object that function is
 *                            called from.
 *  @param[in]    args        Arguments.
 *                args 1:     Statementhandle.
 *                args 2:     Parameter of column number to get data from.
 *  @returns   Tuple.
 *             Tuple[0]: Return code 0 = OK, < 0 if error.
 *             Tuple[1]: The value of the integer.
 *
 *  @remarks   Only the database type INTEGER, INTEGER(n), BIGINT and SMALLINT
 *             may be retrieved using this call.
 */
/* *********************************************************************/
{
  void *statement;
  int rc, column_number;
  int64_t return_value;

  if(!PyArg_ParseTuple(args, "Li", &statement, &column_number)){
    return NULL;
  }

  if(MimerIsNull(statement, column_number) > 0){
    return Py_BuildValue("is", 0, NULL);
  }

  rc = MimerGetInt64(statement, column_number, &return_value);
  return Py_BuildValue("iL", rc, return_value);
}

static PyObject* mimerGetStringC(PyObject* self, PyObject* args)
/* *********************************************************************/
/**
 *  @brief      Gets character data from a result set or output parameter.
 *
 *  @author     Erik Gunne & Magdalena Boström
 *  @date       2017-06-16
 *
 *  @param[in]    self        Pointer to python object that function is
 *                            called from.
 *  @param[in]    args        Arguments.
 *                args 1:     Statementhandle.
 *                args 2:     Parameter of column number to get data from.
 *  @returns   Tuple.
 *             Tuple[0]: Return code >= 0 indicates number of characters in
 *                       column(not counting terminating zero), < 0 if error.
 *             Tuple[1]: The character data.
 *
 *  @remarks   Only SQL data types CHARACTER, CHARACTER VARYING, NATIONAL
 *             CHARACTER, and NATIONAL CHARACTER VARYING may be retrieved
 *             using this function.
 */
/* *********************************************************************/
{
  void *statement;
  int rc, column_number;
  char buf[BUFLEN];

  if(!PyArg_ParseTuple(args, "Li", &statement, &column_number)){
    return NULL;
  }

  if(MimerIsNull(statement, column_number) > 0){
    return Py_BuildValue("is", 0, NULL);
  }

  rc = MimerGetStringC(statement, column_number, buf, BUFLEN);

  if(rc >= BUFLEN) {
    int biglen = rc + 1;
    char *bigbuf = (char*) calloc(biglen, sizeof(char));
    rc = MimerGetStringC(statement, column_number, bigbuf, biglen);
    PyObject* return_object = Py_BuildValue("is", rc, bigbuf);
    free(bigbuf);
    return return_object;
  }
  return Py_BuildValue("is", rc, &buf);
}

static PyObject* mimerGetDouble(PyObject* self, PyObject* args)
/* *********************************************************************/
/**
 *  @brief      Gets double precision float data from a result set or an
 *              output parameter.
 *
 *  @author     Erik Gunne & Magdalena Boström
 *  @date       2017-06-16
 *
 *  @param[in]    self        Pointer to python object that function is
 *                            called from.
 *  @param[in]    args        Arguments.
 *                args 1:     Statementhandle.
 *                args 2:     Parameter of column number to get data from.
 *  @returns   Tuple.
 *             Tuple[0]: Return code 0 = OK, < 0 if error.
 *             Tuple[1]: The double precision value.
 *  @remarks   Only the database data types DOUBLE PRECISION, FLOAT(n), FLOAT
 *             and REAL may be retrieved using this call.
 *
 *  @remarks   Only supported in version 11 or above of the Micro C API.
 */
/* *********************************************************************/
{
  void *statement;
  int rc, column_number;
  double return_value;

  if(!PyArg_ParseTuple(args, "Li", &statement, &column_number)){
    return NULL;
  }

  if(MimerIsNull(statement, column_number) > 0){
    return Py_BuildValue("is", 0, NULL);
  }

  rc = MimerGetDouble(statement, column_number, &return_value);
  return Py_BuildValue("id", rc, return_value);
}

static PyObject* mimerSetInt32(PyObject* self, PyObject* args)
/* *********************************************************************/
/**
 *  @brief      Sets an integer parameter.
 *
 *  @author     Erik Gunne & Magdalena Boström
 *  @date       2017-06-16
 *
 *  @param[in]    self        Pointer to python object that function is
 *                            called from.
 *  @param[in]    args        Arguments.
 *                args 1:     Statementhandle.
 *                args 2:     A number identifying the parameter.
 *                args 3:     An integer value.
 *  @returns   Return code 0 = OK, < 0 if error.
 *
 *  @remarks   Only the database data types INTEGER, INTEGER(n), BIGINT and
 *             SMALLINT may be set using this call.
 */
/* *********************************************************************/
{
  void *statement;
  int rc, parameter_number, value;

  if(!PyArg_ParseTuple(args, "Lii", &statement, &parameter_number, &value)){
    return NULL;
  }
  rc = MimerSetInt32(statement, parameter_number, value);
  return Py_BuildValue("i", rc);
}

static PyObject* mimerSetInt64(PyObject* self, PyObject* args)
/* *********************************************************************/
/**
 *  @brief      Sets an integer parameter.
 *
 *  @author     Erik Gunne & Magdalena Boström
 *  @date       2017-06-16
 *
 *  @param[in]    self        Pointer to python object that function is
 *                            called from.
 *  @param[in]    args        Arguments.
 *                args 1:     Statementhandle.
 *                args 2:     A number identifying the parameter.
 *                args 3:     An integer value.
 *  @returns   Return code 0 = OK, < 0 if error.
 *
 *  @remarks   Only the database data types INTEGER, INTEGER(n), BIGINT and
 *             SMALLINT may be set using this call.
 */
/* *********************************************************************/
{
  void *statement;
  int rc, parameter_number;
  int64_t value;

  if(!PyArg_ParseTuple(args, "LiL", &statement, &parameter_number, &value)){
    return NULL;
  }
    rc = MimerSetInt64(statement, parameter_number, value);
  return Py_BuildValue("i", rc);
}

static PyObject* mimerSetStringC(PyObject* self, PyObject* args)
/* *********************************************************************/
/**
 *  @brief      Sets a string parameter.
 *
 *  @author     Erik Gunne & Magdalena Boström
 *  @date       2017-06-16
 *
 *  @param[in]    self        Pointer to python object that function is
 *                            called from.
 *  @param[in]    args        Arguments.
 *                args 1:     Statementhandle.
 *                args 2:     A number identifying the parameter.
 *                args 3:     An pointer to buffer holding UTF-16 or UTF-32
 *                            character string.
 *  @returns   Return code 0 = OK, < 0 if error.
 *
 *  @remarks   Only the database data types CHARACTER, CHARACTER VARYING,
 *             NATIONAL CHARACTER, and NATIONAL CHARACTER VARYING may be
 *             set using this call.
 */
/* *********************************************************************/
{
  void *statement;
  int rc, parameter_number;
  char* value;

  if(!PyArg_ParseTuple(args, "Lis", &statement, &parameter_number, &value)){
    return NULL;
  }

  rc = MimerSetStringC(statement, parameter_number, value);
  return Py_BuildValue("i", rc);
}

static PyObject* mimerSetDouble(PyObject* self, PyObject* args)
/* *********************************************************************/
/**
 *  @brief      Sets a double floating point parameter.
 *
 *  @author     Erik Gunne & Magdalena Boström
 *  @date       2017-06-16
 *
 *  @param[in]    self        Pointer to python object that function is
 *                            called from.
 *  @param[in]    args        Arguments.
 *                args 1:     Statementhandle.
 *                args 2:     A number identifying the parameter.
 *                args 3:     A double floating point value.
 *  @returns    Return code 0 = OK, < 0 if error.
 *
 *  @remarks    Only supported in version 11 or above of the Micro C API.
 */
/* *********************************************************************/
{
  void *statement;
  int rc, parameter_number;
  double value;
  char *data_null;

  if(!PyArg_ParseTuple(args, "Lid", &statement, &parameter_number, &value,
                        &data_null)){
    return NULL;
  }

  if(data_null == NULL){
    rc = MimerSetNull(statement, parameter_number);
    return Py_BuildValue("i", rc);
  }
  rc = MimerSetDouble(statement, parameter_number, value);

  return Py_BuildValue("i", rc);
}

static PyObject* mimerGetErrorC(PyObject* self, PyObject* args)
{
  /* *********************************************************************/
  /**
   *  @brief      Gets a error message from a mimer session or mimer
   *              statement.
   *
   *  @author     Erik Gunne & Magdalena Boström
   *  @date       2017-06-16
   *
   *  @param[in]    self        Pointer to python object that function is
   *                            called from.
   *  @param[in]    args        Arguments.
   *                args 1:     Statementhandle or sessionhandle.
   *
   *  @returns    Tuple.
   *              Tuple[0]: Return code 0 = OK, < 0 if error.
   *              Tuple[1]: The number of the error found.
   *              Tuple[2]: The message from above error.
   *
   *  @remarks    Only supported in version 11 or above of the Micro C API.
   *              Not all errors are supported by this function.
   */
  /* *********************************************************************/
  void *statement;
  int rc, error_value;
  char message[BUFLEN];

  if(!PyArg_ParseTuple(args, "L", &statement)){
    return NULL;
  }

  rc = MimerGetErrorC(statement, &error_value, message, BUFLEN);

  return Py_BuildValue("iis", rc, error_value, message);
}

/*static PyObject* mimerSetLob(PyObject* self, PyObject* args)
{
  void *statement;
  int length;
  int rc;
  int param_no;
  MimerLob lobhandle;

  if(!PyArg_ParseTuple(args, "Lii", &statement, &param_no, &length)){
    return NULL;
  }

  rc = MimerSetLob(statement, param_no, length, &lobhandle);

  return Py_BuildValue("iL", rc, (int64_t)lobhandle);
}*/

static PyObject* mimerSetBlobData(PyObject* self, PyObject* args)
{
  /* *********************************************************************/
  /**
   *  @brief      Sets a blob data parameter.
   *
   *  @author     Erik Gunne & Magdalena Boström
   *  @date       2017-06-16
   *
   *  @param[in]    self        Pointer to python object that function is
   *                            called from.
   *  @param[in]    args        Arguments.
   *                args 1:     Statementhandle.
   *                args 2:     A number identifying the parameter.
   *                args 3:     A string containing the blob data.
   *  @returns    Return code 0 = OK, < 0 if error.
   *
   *  @remarks    Only supported in version 11 or above of the Micro C API.
   */
  /* *********************************************************************/
  void *statement;
  int rc, parameter_number, parse_length;
  MimerLob lobhandle;
  const char *data;

  // parse_length is a outparameter that returns the lengths of the string data
  if(!PyArg_ParseTuple(args, "Liz#", &statement, &parameter_number, &data,
                        &parse_length)){
    return NULL;
  }

  if(data == NULL){
    rc = MimerSetNull(statement, parameter_number);
    return Py_BuildValue("i", rc);
  }

  rc = MimerSetLob(statement, parameter_number, parse_length, &lobhandle);

  if(rc < 0){
    return Py_BuildValue("i", rc);;
  }

  rc = MimerSetBlobData(&lobhandle, data, parse_length);
  return Py_BuildValue("i", rc);
}

static PyObject* mimerSetNclobDataC(PyObject* self, PyObject* args)
{
  /* *********************************************************************/
  /**
   *  @brief      Sets a nclob data parameter.
   *
   *  @author     Erik Gunne & Magdalena Boström
   *  @date       2017-06-16
   *
   *  @param[in]    self        Pointer to python object that function is
   *                            called from.
   *  @param[in]    args        Arguments.
   *                args 1:     Statementhandle.
   *                args 2:     A number identifying the parameter.
   *                args 3:     A string containing the nclob data.
   *  @returns    Return code 0 = OK, < 0 if error.
   *
   *  @remarks    Only supported in version 11 or above of the Micro C API.
   */
  /* *********************************************************************/
  void *statement;
  int rc, parameter_number, parse_length;
  MimerLob lobhandle;
  const char *data;

  // parse_length is a outparameter that returns the lengths of the string data
  if(!PyArg_ParseTuple(args, "Liz#", &statement, &parameter_number, &data,
                        &parse_length)){
    return NULL;
  }

  if(data == NULL){
    rc = MimerSetNull(statement, parameter_number);
    return Py_BuildValue("i", rc);
  }

  rc = MimerSetLob(statement, parameter_number, parse_length, &lobhandle);

  if(rc < 0){
    return Py_BuildValue("i", rc);;
  }

  rc = MimerSetNclobDataC(&lobhandle, data, parse_length);
  return Py_BuildValue("i", rc);
}


static PyObject* mimerGetBlobData(PyObject* self, PyObject* args)
{
  /* *********************************************************************/
  /**
   *  @brief      Gets a blob data parameter.
   *
   *  @author     Erik Gunne & Magdalena Boström
   *  @date       2017-06-16
   *
   *  @param[in]    self        Pointer to python object that function is
   *                            called from.
   *  @param[in]    args        Arguments.
   *                args 1:     Statementhandle or sessionhandle.
   *                args 2:     A number identifying the parameter.
   *
   *  @returns    Tuple.
   *              Tuple[0]: Return code 0 = OK, < 0 if error.
   *              Tuple[1]: The blob data
   *
   *  @remarks    Only supported in version 11 or above of the Micro C API.
   */
  /* *********************************************************************/
  void *statement;
  int rc, parameter_number;
  MimerLob lobhandle;
  size_t length;

  if(!PyArg_ParseTuple(args, "Li", &statement, &parameter_number)){
    return NULL;
  }

  if(MimerIsNull(statement, parameter_number) > 0){
    return Py_BuildValue("is", 0, NULL);
  }

  rc = MimerGetLob(statement, parameter_number, &length, &lobhandle);

  if(rc < 0){
    return Py_BuildValue("i", rc);;
  }

  char *data = (char*) calloc(length, sizeof(char));
  rc = MimerGetBlobData(&lobhandle, data, length);
  PyObject* return_object = Py_BuildValue("iy#", rc, data,length);
  free(data);
  return return_object;
}

static PyObject* mimerGetNclobDataC(PyObject* self, PyObject* args)
{
  /* *********************************************************************/
  /**
   *  @brief      Gets a nclob data parameter.
   *
   *  @author     Erik Gunne & Magdalena Boström
   *  @date       2017-06-16
   *
   *  @param[in]    self        Pointer to python object that function is
   *                            called from.
   *  @param[in]    args        Arguments.
   *                args 1:     Statementhandle or sessionhandle.
   *                args 2:     A number identifying the parameter.
   *
   *  @returns    Tuple.
   *              Tuple[0]: Return code 0 = OK, < 0 if error.
   *              Tuple[1]: The nclob data
   *
   *  @remarks    Only supported in version 11 or above of the Micro C API.
   */
  /* *********************************************************************/
  void *statement;
  int rc, parameter_number;
  MimerLob lobhandle;
  size_t length, numb_of_bytes;


  if(!PyArg_ParseTuple(args, "Li", &statement, &parameter_number)){
    return NULL;
  }

  if(MimerIsNull(statement, parameter_number) > 0){
    return Py_BuildValue("is", 0, NULL);
  }

  rc = MimerGetLob(statement, parameter_number, &length, &lobhandle);

  if(rc < 0){
    return Py_BuildValue("i", rc);;
  }

  char *data = (char*) calloc(length, sizeof(char));
  rc = MimerGetNclobDataC(&numb_of_bytes, &lobhandle, data, length + 1);
  PyObject* return_object = Py_BuildValue("is", rc, data);
  free(data);
  return return_object;
}

static PyObject* mimerSetBinary(PyObject* self, PyObject* args)
{
  /* *********************************************************************/
  /**
   *  @brief      Sets a binary parameter.
   *
   *  @author     Erik Gunne & Magdalena Boström
   *  @date       2017-06-16
   *
   *  @param[in]    self        Pointer to python object that function is
   *                            called from.
   *  @param[in]    args        Arguments.
   *                args 1:     Statementhandle.
   *                args 2:     A number identifying the parameter.
   *                args 3:     A double floating point value.
   *  @returns    Return code 0 = OK, < 0 if error.
   *
   *  @remarks    Only supported in version 11 or above of the Micro C API.
   */
  /* *********************************************************************/
  void *statement;
  int rc, parameter_number, parse_length;
  char *value;

  if(!PyArg_ParseTuple(args, "Liz#", &statement, &parameter_number, &value,
                        &parse_length)){
    return NULL;
  }

  rc = MimerSetBinary(statement, parameter_number, value, parse_length);
  return Py_BuildValue("i", rc);
}

static PyObject* mimerGetBinary(PyObject* self, PyObject* args)
{
  /* *********************************************************************/
  /**
   *  @brief      Gets a blob data parameter.
   *
   *  @author     Erik Gunne & Magdalena Boström
   *  @date       2017-06-16
   *
   *  @param[in]    self        Pointer to python object that function is
   *                            called from.
   *  @param[in]    args        Arguments.
   *                args 1:     Statementhandle or sessionhandle.
   *                args 2:     A number identifying the parameter.
   *
   *  @returns    Tuple.
   *              Tuple[0]: Return code 0 = OK, < 0 if error.
   *              Tuple[1]: The blob data
   *
   *  @remarks    Only supported in version 11 or above of the Micro C API.
   */
  /* *********************************************************************/
  void *statement;
  int rc, parameter_number;
  char buf[BUFLEN];

  if(!PyArg_ParseTuple(args, "Li", &statement, &parameter_number)){
    return NULL;
  }

  if(MimerIsNull(statement, parameter_number) > 0){
    return Py_BuildValue("is", 0, NULL);
  }

  rc = MimerGetBinary(statement, parameter_number, buf, BUFLEN);

  if(rc >= BUFLEN) {
    int biglen = rc + 1;
    char *bigbuf = (char*) calloc(biglen, sizeof(char));
    rc = MimerGetBinary(statement, parameter_number, bigbuf, biglen);
    PyObject* return_object = Py_BuildValue("iy#", rc, biglen);
    free(bigbuf);
    return return_object;
  }
  return Py_BuildValue("iy#", rc, &buf, rc);
}

static PyObject* mimerSetBoolean(PyObject* self, PyObject* args)
{
  /* *********************************************************************/
  /**
   *  @brief      Sets a boolean parameter.
   *
   *  @author     Erik Gunne & Magdalena Boström
   *  @date       2017-06-16
   *
   *  @param[in]    self        Pointer to python object that function is
   *                            called from.
   *  @param[in]    args        Arguments.
   *                args 1:     Statementhandle.
   *                args 2:     A number identifying the parameter.
   *                args 3:     A boolean value.
   *  @returns    Return code 0 = OK, < 0 if error.
   *
   *  @remarks    In python any object can be a boolean. This function accepts
   *              any python object type in arg 3.
   */
  /* *********************************************************************/
  void *statement;
  int rc, value, parameter_number;

  if(!PyArg_ParseTuple(args, "Lip", &statement, &parameter_number, &value)){
    return NULL;
  }

  rc = MimerSetBoolean(statement, parameter_number, value);
  return Py_BuildValue("i", rc);
}

static PyObject* mimerGetBoolean(PyObject* self, PyObject* args)
{
  /* *********************************************************************/
  /**
   *  @brief      Gets a boolean parameter.
   *
   *  @author     Erik Gunne & Magdalena Boström
   *  @date       2017-06-16
   *
   *  @param[in]    self        Pointer to python object that function is
   *                            called from.
   *  @param[in]    args        Arguments.
   *                args 1:     Statementhandle.
   *                args 2:     A number identifying the parameter.
   *
   *  @returns    Tuple.
   *              Tuple[0]: Return code 0 = OK, < 0 if error.
   *              Tuple[1]: Same as Tuple[0].
   *
   *  @remarks    Py_BuildValue does not support returning a python bool object.
   *              The reutrn value is given as a int bool(0 == ture, 1 == false).
   */
  /* *********************************************************************/
  void *statement;
  int rc, parameter_number;

  if(!PyArg_ParseTuple(args, "Li", &statement, &parameter_number)){
    return NULL;
  }
  if(MimerIsNull(statement, parameter_number) > 0){
    return Py_BuildValue("ii", 0, 0);
  }

  rc = MimerGetBoolean(statement, parameter_number);

  return Py_BuildValue("ii", rc,rc);
}

static PyObject* mimerSetNull(PyObject* self, PyObject* args)
{
  /* *********************************************************************/
  /**
   *  @brief      Sets a parameter to sql NULL.
   *
   *  @author     Erik Gunne & Magdalena Boström
   *  @date       2017-06-16
   *
   *  @param[in]    self        Pointer to python object that function is
   *                            called from.
   *  @param[in]    args        Arguments.
   *                args 1:     Statementhandle or sessionhandle.
   *                args 2:     A number identifying the parameter.
   *
   *  @returns    Return code 0 = OK, < 0 if error.
   *
   */
  /* *********************************************************************/
  void *statement;
  int rc, parameter_number;
  char *nullValue;

  if(!PyArg_ParseTuple(args, "Liz", &statement, &parameter_number, &nullValue)){
    return NULL;
  }

  rc = MimerSetNull(statement, parameter_number);
  return Py_BuildValue("i", rc);
}

static PyObject* mimerColumnNameC(PyObject* self, PyObject* args)
{
  /* *********************************************************************/
  /**
   *  @brief      Obtains the name of a column in a multibyte character
   *              string.
   *
   *  @author     Erik Gunne & Magdalena Boström
   *  @date       2017-07-20
   *
   *  @param[in]    self        Pointer to python object that function is
   *                            called from.
   *  @param[in]    args        Arguments.
   *                args 1:     Statementhandle.
   *                args 2:     A number identifying the column number.
   *
   *  @returns    Return code 0 >= Length of column name, < 0 if error.
   *
   */
  /* *********************************************************************/
  void *statement;
  int rc, column_number;
  char buf[BUFLEN];

  if(!PyArg_ParseTuple(args, "Li", &statement, &column_number)){
    return NULL;
  }

  rc = MimerColumnNameC(statement, column_number, buf, BUFLEN);

  if(rc >= BUFLEN) {
    int biglen = rc + 1;
    char *bigbuf = (char*) calloc(biglen, sizeof(char));
    rc = MimerColumnNameC(statement, column_number, bigbuf, biglen);
    PyObject* return_object = Py_BuildValue("is", rc, bigbuf);
    free(bigbuf);
    return return_object;
  }

  return Py_BuildValue("is", rc, &buf);
}
/**
* Describes all methods of cursor extension.
*/
static PyMethodDef cursorMethods[] =
{
  {"mimerBeginStatementC", (PyCFunction)mimerBeginStatementC, METH_VARARGS,
    "Prepares an SQL statement for execution"},
  {"mimerEndStatement", (PyCFunction)mimerEndStatement, METH_VARARGS,
    "Closes a prepared statement."},
  {"mimerOpenCursor", (PyCFunction)mimerOpenCursor, METH_VARARGS,
    "Opens a cursor."},
  {"mimerCloseCursor", (PyCFunction)mimerCloseCursor, METH_VARARGS,
    "Closes an open cursor."},
  {"mimerExecuteStatementC", (PyCFunction)mimerExecuteStatementC, METH_VARARGS,
    "Executes a statement directly without parameters."},
  {"mimerExecute", (PyCFunction)mimerExecute, METH_VARARGS,
    "Executes a statement that does not return a result set."},
  {"mimerParameterCount", (PyCFunction)mimerParameterCount, METH_VARARGS,
    "Returns the number of parameters of a statement."},
  {"mimerParameterType", (PyCFunction)mimerParameterType, METH_VARARGS,
    "Obtains the data type of a pararmeter"},
  {"mimerColumnCount", (PyCFunction)mimerColumnCount, METH_VARARGS,
    "Obtains the number of columns in a result set."},
  {"mimerColumnType", (PyCFunction)mimerColumnType, METH_VARARGS,
    "Returns the type of a column."},
  {"mimerFetch", (PyCFunction)mimerFetch, METH_VARARGS,
    "Advances to the next row of the result set."},
  {"mimerGetInt32", (PyCFunction)mimerGetInt32, METH_VARARGS,
    "Get int32 data from a result set or output pararmeter."},
  {"mimerGetInt64", (PyCFunction)mimerGetInt64, METH_VARARGS,
    "Get int64 data from a result set or output pararmeter."},
  {"mimerGetStringC", (PyCFunction)mimerGetStringC, METH_VARARGS,
    "Get character data from a result set or output pararmeter."},
  {"mimerGetDouble", (PyCFunction)mimerGetDouble, METH_VARARGS,
    "Get double precision float data from a result set or output pararmeter."},
  {"mimerSetInt32", (PyCFunction)mimerSetInt32, METH_VARARGS,
    "Sets an int32 parameter."},
  {"mimerSetInt64", (PyCFunction)mimerSetInt64, METH_VARARGS,
    "Sets an int64 parameter."},
  {"mimerSetStringC", (PyCFunction)mimerSetStringC, METH_VARARGS,
    "Sets an string parameter."},
  {"mimerSetDouble", (PyCFunction)mimerSetDouble, METH_VARARGS,
    "Sets an double folating point parameter."},
  {"mimerGetErrorC", (PyCFunction)mimerGetErrorC, METH_VARARGS,
    "Returns error message."},
  {"mimerSetBlobData", (PyCFunction)mimerSetBlobData, METH_VARARGS,
  "Sets blob data."},
  {"mimerGetBlobData", (PyCFunction)mimerGetBlobData, METH_VARARGS,
  "Gets blob data."},
  {"mimerSetNclobDataC", (PyCFunction)mimerSetNclobDataC, METH_VARARGS,
  "sets Nclob data."},
  {"mimerGetNclobDataC", (PyCFunction)mimerGetNclobDataC, METH_VARARGS,
  "Gets Nclob data."},
  {"mimerSetBinary", (PyCFunction)mimerSetBinary, METH_VARARGS,
  "Sets binary data."},
  {"mimerSetBoolean", (PyCFunction)mimerSetBoolean, METH_VARARGS,
  "Sets boolean data."},
  {"mimerGetBoolean", (PyCFunction)mimerGetBoolean, METH_VARARGS,
  "Gets boolean data."},
  {"mimerSetNull", (PyCFunction)mimerSetNull, METH_VARARGS,
  "Sets parameter to null."},
  {"mimerColumnNameC", (PyCFunction)mimerColumnNameC, METH_VARARGS,
  "Obtains name of a column."},
  {"mimerGetBinary", (PyCFunction)mimerGetBinary, METH_VARARGS,
  "Obtains name of a column."},
  {NULL, NULL, 0, NULL}
};

/**
* Module defintion struct.
*/
static struct PyModuleDef cursor = {
  PyModuleDef_HEAD_INIT,
  "cursor",
  "cursor module",
  -1,
  cursorMethods
};

PyMODINIT_FUNC PyInit_cursor(void)
{
     return PyModule_Create(&cursor);
}
